/**
 * Created by chenrozenes and ofirhartuv on 05/11/2017.
 */
import { Client } from 'pg';
import moment from 'moment';
import Promise from 'bluebird';

import logger from '../logger';
import Puller from './puller';

const TABLE_NAME = 'rabbit_pull_queue_sync';
const CREATE_QUERY = `CREATE TABLE IF NOT EXISTS ${TABLE_NAME} (` +
  'queue CHARACTER VARYING(200) NOT NULL UNIQUE, ' +
  'last_pull TIMESTAMP NOT NULL);';

const INIT_FIRST_VALUE = `INSERT INTO ${TABLE_NAME} VALUES($1,'-infinity') ON CONFLICT DO NOTHING`;

const UPDATE_QUERY = `UPDATE ${TABLE_NAME} SET last_pull=$1 WHERE queue LIKE $2 AND last_pull<$3`;

/**
 * Puller class that uses Postgres database in order to sync pulling between different workers
 */
class PostgresPuller extends Puller {
  constructor(config = {}) {
    super(config);
    this.connStr = config.options.url;
  }

  _connect() {
    logger.instance.info(`[AMQP] PostgresPuller - connecting to Postgres: ${this.connStr}`);
    this.client = new Client(this.connStr);
    return Promise.resolve()
      .then(() => this.client.connect());
  }

  _disconnect() {
    logger.instance.info('[AMQP] PostgresPuller - closing connection');
    return Promise.resolve()
      .then(() => this.client.end());
  }

  /**
   * Creates the table and fills with initial values if not exist, before starting pulling loop
   * @param queue
   * @returns {*}
   * @private
   */
  _setup(queue) {
    logger.instance.info('[AMQP] PostgresPuller - Creating sync table with initial values');
    return Promise.resolve()
      .then(() => this.client.query(CREATE_QUERY))
      .then(() => this.client.query(INIT_FIRST_VALUE, [queue]));
  }

  /**
   * Trying to update the queue row, if succeeded current worker is allowed to pull.
   *
   * @returns {*}
   */
  shouldPull() {
    const now = moment();
    const lessThen = moment(now).subtract(this.interval / 1000, 'seconds'); // Subtract interval

    return Promise.resolve()
      .then(() => this.client.query(UPDATE_QUERY, [
        now.toDate(),
        this.queue,
        lessThen.toDate(),
      ])).then((result) => {
        // More than one row updated is weird
        if (result.rowCount > 1) {
          logger.instance.error(`[AMQP] Puller updated ${result.rowCount} row. ` +
            `Syncing queue ${this.queue} is problematic`);

          throw new Error('PostgresPuller performed bad update');
        }

        // If updated, it means the worker can pull
        return result.rowCount === 1;
      });
  }

  start(channel, queue, callback) {
    return this._connect()
      .then(() => this._setup(queue))
      .then(() => super.start(channel, queue, callback));
  }

  stop() {
    return super.stop()
      .then(() => this._disconnect());
  }
}

export default PostgresPuller;
